package com.aciedge.cotplitool.cot;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.ConnectException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.Socket;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jdom2.Attribute;
import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;

/**
 * UDP Multicast or TCP CoT PLI message generator/server
 *
 * @author aciedge.com
 * @author
 * <a href="mailto:tdnorbra@nps.edu?subject=com.aciedge.cotplitool.cot.CoTPLISender">Terry Norbraten, NPS MOVES</a>
 */
public class CoTPLISender {

    private final String mIPAddr;
    private final int mPort;

    private String protocol;
    private MulticastSocket mSocket;
    private Socket socket;
    private PrintWriter writer;
    private DatagramPacket pack;
    private ByteArrayOutputStream out;
    private final XMLOutputter xmlOutput;

    private Document cotMsg;

    public CoTPLISender(String protocol, String destinationIP, int port) throws IOException {

        this.protocol = protocol;
        this.mIPAddr = destinationIP;
        this.mPort = port;
        xmlOutput = new XMLOutputter(Format.getPrettyFormat());
        out = new ByteArrayOutputStream();

        switch (protocol) {
            case "udp" -> {
                this.mSocket = new MulticastSocket();
                this.mSocket.setTimeToLive(64);
                pack = new DatagramPacket(new byte[1024], 1024, InetAddress.getByName(this.mIPAddr), this.mPort);
            }
            case "tcp" -> {
                try {
                    socket = new Socket(destinationIP, port);

                    if (socket != null) {
                        OutputStream output = socket.getOutputStream();
                        writer = new PrintWriter(output, true);
                    }
                } catch (ConnectException ex) {
                    Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
            default -> {
                System.err.println("Unknown protocol: " + protocol + ". Defaulting to UDP Multicast");
                this.protocol = "upd";
                this.mSocket = new MulticastSocket();
                this.mSocket.setTimeToLive(64);
                pack = new DatagramPacket(new byte[1024], 1024, InetAddress.getByName(this.mIPAddr), this.mPort);
            }
        }

        buildCotMsg();
    }

    public void sendPLI(String uid, String time, String start, String stale, String lat, String lon, String callsign) {
        String cot = null;
        try {
            cot = this.getCotMsg(uid, time, start, stale, lat, lon, callsign);
        } catch (Exception var9) {
            Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, var9);
        }

        this.sendPacket(cot);
    }

    public void close() {
        try {
            switch (protocol) {
                case "udp" -> {
                    this.mSocket.close();
                }
                case "tcp" -> {
                    if (writer != null) {
                        writer.close(); // flushes buffer contents to the client
                    }
                    if (socket != null) {
                        socket.close();
                    }
                }
            }
        } catch (IOException var2) {
            Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, var2);
        }
    }

    private void buildCotMsg() {
        try (InputStream inStream = getClass().getClassLoader().getResourceAsStream("cotMsgTemplate.xml")) {
            SAXBuilder builder = new SAXBuilder();
            cotMsg = builder.build(inStream);
        } catch (JDOMException | IOException ex) {
            Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private void sendPacket(String msg) {
        try {
            switch (this.protocol) {
                case "udp" -> {
                    pack.setData(msg.getBytes());
                    pack.setLength(msg.length());
                    this.mSocket.send(pack); // individual CoT messages
                }
                case "tcp" -> {
                    if (writer != null) {
                        writer.write(msg); // a whole string of CoT messages
                    }
                }
            }
        } catch (IOException var5) {
            Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, var5);
        }
    }

    private String getCotMsg(String uid, String time, String start, String stale, String lat, String lon, String callsign) {
       
        // <event>
        Element _event = cotMsg.getRootElement();

        Attribute _uid = _event.getAttribute("uid");
        _uid.setValue(uid);
        Attribute _time = _event.getAttribute("time");
        _time.setValue(time);
        Attribute _start = _event.getAttribute("start");
        _start.setValue(start);
        Attribute _stale = _event.getAttribute("stale");
        _stale.setValue(stale);

        // <event>
        //   <point>
        Element _point = _event.getChild("point");

        Attribute _lat = _point.getAttribute("lat");
        _lat.setValue(lat);
        Attribute _lon = _point.getAttribute("lon");
        _lon.setValue(lon);

        // <event>
        //   <point>
        //   <detail>
        Element _detail = _event.getChild("detail");

        // <event>
        //   <point>
        //   <detail>
        //     <contact>
        Element _contact = _detail.getChild("contact");
        Attribute _callsign = _contact.getAttribute("callsign");
        _callsign.setValue(callsign);

        // send xml
        try {
            xmlOutput.output(cotMsg, out);
        } catch (IOException ex) {
            Logger.getLogger(CoTPLISender.class.getName()).log(Level.SEVERE, null, ex);
        }
        String send = out.toString();
        out.reset();
        return send;
    }

} // end class CoTPLISender
