/*
Copyright (c) 1995-2024 held by the author(s).  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials provided with the
      distribution.
    * Neither the names of the Naval Postgraduate School (NPS)
      Modeling, Virtual Environments and Simulation (MOVES) Institute
      (https://www.nps.edu and https://my.nps.edu/web/moves)
      nor the names of its contributors may be used to endorse or
      promote products derived from this software without specific
      prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
package edu.nps.moves.cot.dis;

//import edu.nps.moves.dis7.entities.swe.platform.surface._002Triton;
import edu.nps.moves.dis7.entities.usa.platform.air.ScanEagleA15;
import edu.nps.moves.dis7.enumerations.*;
import edu.nps.moves.dis7.pdus.*;
import edu.nps.moves.dis7.utilities.CoordinateConversions;
import edu.nps.moves.dis7.utilities.DisChannel;
import edu.nps.moves.dis7.utilities.PduFactory;

import java.io.StringReader;
import java.nio.charset.Charset;
import java.time.LocalDateTime;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.bind.JAXBException;

import mil.af.cursorOnTarget.CotDetail;
import mil.af.cursorOnTarget.CotEvent;
import mil.af.cursorOnTarget.CotPoint;
import mil.af.cursorOnTarget.DetailSubelement;
import mil.af.cursorOnTarget.UnsupportedCotVersionException;

import mil.army.usareur.g3.mcsd.CoTUtils.CoTparser;
import mil.army.usareur.g3.mcsd.CotBindings.Spatial;
import mil.army.usareur.g3.mcsd.CotBindings.Spatial.Attitude;
import mil.army.usareur.g3.mcsd.CotBindings.Track;

/** Cursor on Target (CoT) received message parser and transmitter of DIS PDUs
 * populated with CoT data. Default program initialization includes PDU 
 * recording turned on by default.
 * 
 * Based on: NetworkedGraphicsMV3500/examples/src/OpenDis7Examples/ExampleSimulationProgram.java
 *
 * @author <a href="mailto:tdnorbra@nps.edu?subject=edu.nps.moves.cot.dis.CoTMsgParserDisRunner">Terry Norbraten, NPS MOVES</a>
 */
public class CoTMsgParserDisRunner extends CoTparser
{
    /* **************************** infrastructure code, modification is seldom needed ************************* */
                 
    private   String     descriptor;
    /** DIS channel defined by network address/port combination includes multiple utility capabilities */
    private DisChannel disChannel;
    /** Factory object used to create new PDU instances */
    private PduFactory pduFactory;
    
    /** seconds per loop for real-time or simulation execution */
    private double  simulationTimeStepDuration; // seconds TODO encapsulate
    /** initial simulation time in seconds */
    private double  simulationTimeInitial;
    /** current simulation time in seconds */
    private double  simulationTimeSeconds;
    /** Maximum number of simulation loops */
    private int MAX_LOOP_COUNT;
    
    private String narrativeMessage1;
    private String narrativeMessage2;
    private String narrativeMessage3;
    
    /** ESPDU for entity 1 */
    private EntityStatePdu     entityStatePdu_1;
    /** ESPDU for entity 2 */
//    private EntityStatePdu     entityStatePdu_2;
    /** FirePdu for entity 1 first  weapon (if any) */
    private FirePdu            firePdu_1a;
    /** FirePdu for entity 1 second weapon (if any) */
//    private FirePdu            firePdu_1b;
    /** MunitionDescriptor for these weapons */
//    private MunitionDescriptor munitionDescriptor1;
    
    // hey programmer, what other state do you want?  this is a good place to declare it...
    
    /**
     * Constructor to create an instance of this class.
     * Design goal: additional built-in initialization conveniences can go here
     * to keep your efforts focused on the runSimulation() method.
     */
    public CoTMsgParserDisRunner()
    {
        descriptor = this.getClass().getSimpleName();
        simulationTimeStepDuration = 1.0d;
        simulationTimeSeconds = simulationTimeInitial = 0.0d;
        MAX_LOOP_COUNT = 4;
        initialize();
    }
    /**
     * Constructor to create an instance of this class.
     * @param newDescriptor describes this program, useful for logging and debugging
     */
    public CoTMsgParserDisRunner(String newDescriptor)
    {   
        this();
        descriptor = newDescriptor;
    }
    /**
     * Utility Constructor that allows your example simulation program to override default network address and port
     * @param address network address to use
     * @param port corresponding network port to use
     */
    public CoTMsgParserDisRunner(String address, int port)
    {
        this();
        disChannel.setNetworkAddress            (address);
        disChannel.setNetworkPort               (port);
    }
    
    /** Initialize channel setup for OpenDis7 and report a test PDU
     * @see initializeDisChannel
     * @see initializeSimulationEntities 
     */
    private void initialize()
    {
        initializeDisChannel(); // must come first, uses PduFactory
        
        initializeSimulationEntities(); // set unchanging parameters
        
        getDisChannel().join(); // TODO further functionality expected
        
//        String timeStepMessage = "Simulation timestep duration " + getSimulationTimeStepDuration() + " seconds";
//        getDisChannel().sendCommentPdu(simulationTimeSeconds, DisChannel.COMMENTPDU_SIMULATION_TIMESTEP, timeStepMessage);
        // additional constructor initialization can go here
    }
    
    /** Initialize channel setup for OpenDis7 and report a test PDU */
    private void initializeDisChannel()
    {
        if (getDisChannel() == null)
            disChannel = new DisChannel();
        else
        {
            getDisChannel().printlnTRACE ("*** warning, duplicate invocation of initializeDisChannel() ignored");
            return;
        }
        pduFactory     = getDisChannel().getPduFactory();
        getDisChannel().setDescriptor(descriptor); // ExampleSimulationProgram might be a superclass
        getDisChannel().setUpNetworkInterface();

        getDisChannel().setVerboseComments           (false);       // TODO rename library method to disambiguate CommentPDU
                                                                                   // TODO still seems really chatty... add silent mode?
        getDisChannel().setVerboseDisNetworkInterface(false); // Default false
        getDisChannel().getPduRecorder().setVerbose  (true);
        // TODO confirm whether recorder is explicitly started by programmer (or not)
    }
    
    /** Get ready, get set... initialize simulation entities.  Who's who in the zoo? */
    private void initializeSimulationEntities()
    {        
        if (pduFactory == null)
            pduFactory      = getDisChannel().getPduFactory();
        entityStatePdu_1    = pduFactory.makeEntityStatePdu();
//        entityStatePdu_2    = pduFactory.makeEntityStatePdu();
//        firePdu_1a          = pduFactory.makeFirePdu();
//        firePdu_1b          = pduFactory.makeFirePdu();
//        munitionDescriptor1 = new MunitionDescriptor();
        
        // Your model setup: define participants.  who's who in this zoo?
        // Assuming you keep track of entity objects...  here is some support for for Entity 1.
        
        // PDU objects are already declared and instances created, so now set their values.
        
        /***** Initialize each PDU to a known state *****/
        entityStatePdu_1.getEntityID().setSiteID(1).setApplicationID(2).setEntityID(3); // made-up example ID;
        getDisChannel().addEntity(entityStatePdu_1.getEntityID());
        // TODO someday, use enumerations for sites as part of a SimulationManager object; e.g. is there a unique site triplet for MOVES Institute?

        // Essential that we have a known initial state for DeadReckoningParameters
        DeadReckoningParameters drp = entityStatePdu_1.getDeadReckoningParameters();

        // Set Dead Reckoning Model (DRM), Rotational Velocity in World Coordinates (RVW)
        drp.setDeadReckoningAlgorithm(DeadReckoningAlgorithm.DRM_RVW_HIGH_SPEED_OR_MANEUVERING_ENTITY_WITH_EXTRAPOLATION_OF_ORIENTATION);
        /***** End Initialize pdu *****/
        
        /***** Begin vehicle specific DIS Enumerations *****/
        entityStatePdu_1.setForceId(ForceID.FRIENDLY);
        entityStatePdu_1.setEntityType(new ScanEagleA15()); // note import statement above
        EntityType type = entityStatePdu_1.getEntityType();

        // Use ASCII for the character set
        entityStatePdu_1.getMarking().setCharacterSet(EntityMarkingCharacterSet.ASCII);
        Charset cs = Charset.forName("US-ASCII");
        byte[] charArray = cs.encode(type.toString() + entityStatePdu_1.getEntityID()).array();
        entityStatePdu_1.getMarking().setCharacters(charArray);

        type.setEntityKind(EntityKind.PLATFORM);
        type.setDomain(Domain.inst(PlatformDomain.AIR));

        // Zero until we know more about how we want to define capabilities
        entityStatePdu_1.setCapabilities(new AirPlatformCapabilities());

        // Need to set bits 1-32 for proper initialization
        entityStatePdu_1.setEntityAppearance(0);

        // Side number of SubCategory
        type.setSpecific(0);

        // Weapons loading?
        type.setExtra((short) 0);
        /***** End vehicle specific DIS Enumerations *****/

//        entityStatePdu_2.getEntityID().setSiteID(1).setApplicationID(2).setEntityID(4); // made-up example ID; 
//        getDisChannel().addEntity(entityStatePdu_2.getEntityID());
//        entityStatePdu_2.setForceId(ForceID.OPPOSING);
//        entityStatePdu_2.setEntityType(new _002Triton()); // note import statement above
//        entityStatePdu_2.setMarking("Entity #2");

        // TODO how should we customize this munition?  what are key parameters for your simulation? 
        // more is needed here by scenario authors...
//        munitionDescriptor1.setQuantity(1);
//        firePdu_1a.setDescriptor(munitionDescriptor1).setRange(1000.0f);

        simulationTimeInitial = entityStatePdu_1.getTimestamp();
    }
                 
    /**
     * This runSimulationLoops() method is for you, a customizable programmer-modifiable
     * code block for defining and running a new simulation of interest.
     * 
     * Welcome! Other parts of this program handle bookkeeping and plumbing tasks so that
     * you can focus on your model entities and activities.
     * Expandable support includes DIS EntityStatePdu, FirePdu and CommentPdu all available for 
     * modification and sending in a simulation loop.
     * Continuous improvement efforts seek to make this program as easy and straightforward
     * as possible for DIS simulationists to use and adapt.
     * All of the other methods are setup, teardown and configuration that you may find
     * interesting, even helpful, but don't really have to worry about.
     */
    @SuppressWarnings("SleepWhileInLoop") // yes we might do that
    public void runSimulationLoops ()
    {
      try
      {              
        final int     SIMULATION_MAX_LOOP_COUNT = 10; // be deliberate out there!  also avoid infinite loops.
              int     simulationLoopCount = 0;        // variable, initialized at 0
              boolean simulationComplete = false;     // sentinel variable as termination condition, are we done yet?
        
        // TODO reset Clock Time for today's date and timestamp to zero, providing consistent outputs for each simulation run
        String timeMessage = "Simulation time " + simulationTimeSeconds + " at LocalDateTime " + LocalDateTime.now();
            getDisChannel().sendCommentPdu(simulationTimeSeconds, DisChannel.COMMENTPDU_TIME, timeMessage);
        // TODO replace enumeration with disChannel.COMMENTPDU_TIME
        // TODO fix VariableRecordType.TIME_AMP_DATE_VALID
        
        // ===================================================================================================
        // loop the simulation while allowed, programmer can set additional conditions to break out and finish
        while (simulationLoopCount < SIMULATION_MAX_LOOP_COUNT)  // are we done yet?
        {
            simulationLoopCount++; // good practice: increment loop counter as first action in that loop
            
            // =============================================================================================
            // * your own simulation code starts here! *****************************************************
            // =============================================================================================
            
            //  are there any other variables to modify at the beginning of your loop?
            
            // are your reading any DIS PDUs from the network?  check for them here
            
            // compute a track, update an ESPDU, whatever it is that your model is doing...
            
            // Where is my entity?  Insert changes in position; this sample only changes X position.
            entityStatePdu_1.getEntityLocation().setX(entityStatePdu_1.getEntityLocation().getX() + 1.0); // 1m per timestep
            
            // decide whether to fire, and then update the firePdu.  Hmmm, you might want a target to shoot at!
            
            // etc. etc. your code goes here for your simulation of interest
                
            // something happens between my simulation entities, la de da de da...
            System.out.println ("... My simulation just did something, no really...");
            System.out.flush(); // make sure this arrives to user even if other threads somehow become deadlocked
            
            
            // make your reports: narrative code for CommentPdu here (set all to empty strings to avoid sending)
            narrativeMessage1 = "MV3500 ExampleSimulationProgram";
            narrativeMessage2 = "runSimulation() loop " + simulationLoopCount;
            narrativeMessage3 = ""; // intentionally blank for testing

            // your loop termination condition goes here
            if (simulationLoopCount > MAX_LOOP_COUNT) // for example
            {
                simulationComplete = true;
            }      
            // =============================================================================================
            // * your own simulation code is finished here! ************************************************
            // =============================================================================================
            
            // staying synchronized with timestep: wait duration for elapsed time in this loop
            // Thread.sleep needs a (long) parameter for milliseconds, which are clumsy to use sometimes
            Thread.sleep((long)(getSimulationTimeStepDuration() * 1000)); // units of seconds * (1000 msec/sec) = milliseconds
            System.out.println ("... [Pausing for " + getSimulationTimeStepDuration() + " seconds]");
            
            // OK now send the status PDUs for this loop, and then continue
            System.out.println ("... sending PDUs of interest for simulation step " + simulationLoopCount + ", monitor loopback to confirm sent");
            System.out.flush();
            
            // TODO set timesteps in PDUs
            
            sendAllPdusForLoopTimestep(simulationTimeSeconds, entityStatePdu_1, 
                                            firePdu_1a, 
                                         DisChannel.COMMENTPDU_APPLICATION_STATUS, 
                                           narrativeMessage1, narrativeMessage2, narrativeMessage3);
//                getDisChannel().sendSinglePdu(simulationTimeSeconds, entityStatePdu_2); // me too i.e. 2!
            
            System.out.println ("... [PDUs of interest successfully sent for this loop]");
            System.out.flush();
            
            // ===============================
            // current loop now finished, check whether to terminate if simulation complete, otherwise continue
            if (simulationComplete || (simulationLoopCount > 10000)) // for example; including fail-safe condition is good
            {
                System.out.println ("... [loop termination condition met, simulationComplete=" + simulationComplete + "]"); // ", final loopCount=" + loopCount + 
                System.out.flush();
                break;
            }
            simulationTimeSeconds += getSimulationTimeStepDuration(); // good practice: increment simulationTime as lastst action in that loop
                    
        }   // end of simulation loop, continue until done
        // ===================================================================================================// ===================================================================================================// ===================================================================================================// ===================================================================================================

        narrativeMessage2 = "runSimulation() completed successfully"; // all done, so tell everyone else on the channel
        // TODO better javadoc needs to be autogenerated for VariableRecordType enumerations
            getDisChannel().sendCommentPdu(DisChannel.COMMENTPDU_NARRATIVE, narrativeMessage1, narrativeMessage2, narrativeMessage3);
        System.out.println ("... [final=completion CommentPdu successfully sent for simulation]");
        
        getDisChannel().leave(); // embedded SimulationManager is expected to send appropriate PDUs for entity, application shutdown <- Does nothing as of 7/17/24
      } 
      catch (InterruptedException iex) // handle any exception that your code might choose to provoke!
      {
        Logger.getLogger(CoTMsgParserDisRunner.class.getSimpleName()).log(Level.SEVERE, null, iex);
      }
    }

    /**
     * Send EntityState, Fire, Comment PDUs that got updated for this loop, reflecting state of current simulation timestep.
     * @param simTimeSeconds simulation time in second, applied to PDU as timestamp
     * @param entityStatePdu the ESPDU to send, if any
     * @param firePdu        the FirePDU to send, if any
     * @param commentType    enumeration value describing purpose of the narrative comment PDU
     * @param comments       String array of narrative comments
     * @see DisChannel
//   * @see DisTime // TODO find renamed version
     * @see <a href="https://docs.oracle.com/javase/tutorial/java/javaOO/arguments.html">Passing Information to a Method or a Constructor</a> Arbitrary Number of Arguments
     */
    public void sendAllPdusForLoopTimestep(double simTimeSeconds,
                                   EntityStatePdu entityStatePdu,
                                          FirePdu firePdu,
                               VariableRecordType commentType,
                                     // vararg... variable-length set of String comments can optionally follow
                                        String... comments)
    {
        if (entityStatePdu != null)
            getDisChannel().sendSinglePdu(simTimeSeconds, entityStatePdu);
            
        if (firePdu != null)
            getDisChannel().sendSinglePdu(simTimeSeconds, firePdu); // bang
        
        getDisChannel().sendCommentPdu(simTimeSeconds, commentType, comments); // empty comments are filtered
    }
    
    /**
     * Initial execution via main() method: handle args array of command-line initialization (CLI) arguments here
     * @param args command-line parameters: network address and port
     */
    private void handleArguments(String[] args)
    {
        // initial execution: handle args array of initialization arguments here
        if (args.length == 2) 
        {
            if ((args[0] != null) && !args[0].isEmpty())
                thisProgram.getDisChannel().setNetworkAddress(args[0]);
            if ((args[1] != null) && !args[1].isEmpty())
                thisProgram.getDisChannel().setNetworkPort(Integer.parseInt(args[1]));
        }
        else if (args.length != 0) 
        {
            System.err.println("Usage: " + thisProgram.getClass().getSimpleName() + " [address port]");
            System.exit(-1);
        }
    }

    /**
     * Get simple descriptor (such as parent class name) for this network interface, used in trace statements
     * @return simple descriptor name
     */
    public String getDescriptor() {
        return descriptor;
    }

    /**
     * Set new simple descriptor (such as parent class name) for this network interface, used in trace statements
     * @param newDescriptor simple descriptor name for this interface
     */
    public void setDescriptor(String newDescriptor) {
        if (newDescriptor == null)
            newDescriptor = "";
        this.descriptor = newDescriptor;
    }

    /**
     * parameter accessor method
     * @return the simulationTimeStepDuration in seconds
     */
    public double getSimulationTimeStepDuration() {
        return simulationTimeStepDuration;
    }

    /**
     * parameter accessor method
     * @param timeStepDurationSeconds the simulationTimeStepDuration in seconds to set
     */
    public void setSimulationTimeStepDuration(double timeStepDurationSeconds) {
        this.simulationTimeStepDuration = timeStepDurationSeconds;
    }
    
    @Override
    public void coTeventHandler(String xmlEvent) {
//        System.out.println(Constants.CUSTOM_HANDLER_COT_PDU + " EVENT\n" + xmlEvent.trim());
//        System.out.println(Constants.CUSTOM_HANDLER_COT_PDU + " EVENT\n");

        StringReader xmlReader;
        Spatial spa;
        Attitude att;
        Track track;

        double lat, lon, ele;
        
        // https://phys.libretexts.org/Bookshelves/Classical_Mechanics/Variational_Principles_in_Classical_Mechanics_(Cline)/13%3A_Rigid-body_Rotation/13.13%3A_Euler_Angles
        // Euler angles{𝜙,𝜃,𝜓} = phi, theta, psi
        double phi, theta, psi; // h,r,p
        Vector3Double xyz = new Vector3Double(),
              phiThetaPsi = new Vector3Double();
        Vector3Float lv   = new Vector3Float();
        
        try {
        
            CotEvent cotEvent = CotEvent.parse(xmlEvent.trim());
            Date time = cotEvent.getTime();
            String timeStepMessage = "CoT message timestamp: " + time;
            getDisChannel().sendCommentPdu(time.getTime()/1000L, DisChannel.COMMENTPDU_TIME, timeStepMessage);
            entityStatePdu_1.setTimestamp((int) (time.getTime()/1000L)); // milli to sec past 01 JAN 1970 00:00:00 GMT
            
            CotPoint point = cotEvent.getPoint();

            // compute a track, update an ESPDU, whatever it is that your model is doing...
            lat = point.getLatitude();
            lon = point.getLongitude();
            ele = point.getHAE();

            // CoordinateConversions from OpenDIS7 java lib
            xyz.setX(CoordinateConversions.getXYZfromLatLonDegrees(lat, lon, ele)[0])
               .setY(CoordinateConversions.getXYZfromLatLonDegrees(lat, lon, ele)[1])
               .setZ(CoordinateConversions.getXYZfromLatLonDegrees(lat, lon, ele)[2]);

            Vector3Double location = entityStatePdu_1.getEntityLocation();
            location.setX(xyz.getX())
                    .setY(xyz.getY())
                    .setZ(xyz.getZ());

            CotDetail detail = cotEvent.getDetail();
            for (DetailSubelement subElement : detail) {
                xmlReader = new StringReader(subElement.toXml());
                if ("spatial".equalsIgnoreCase(subElement.getName())) {
                    spa = (Spatial) jaxbUnmarshallerTrack.unmarshal(xmlReader);
                    att = spa.getAttitude();
                    
                    phi = att.getYaw(); // hdg
                    theta = att.getRoll();
                    psi = att.getPitch();

                    phi =   Math.toRadians(phi);
                    theta = Math.toRadians(theta);
                    psi =   Math.toRadians(psi);

                    phiThetaPsi.setX(phi).setY(theta).setZ(psi); // h,r,p
                }//spatial
                
                if ("track".equalsIgnoreCase(subElement.getName())) {
                    track = (Track) jaxbUnmarshallerTrack.unmarshal(xmlReader);
                    lv.setX((float) track.getSpeed());
                }//track
                xmlReader.close();
            }

            EulerAngles orientation = entityStatePdu_1.getEntityOrientation();
            orientation.setPhi((float) phiThetaPsi.getX())  // hdg -> phi
                     .setTheta((float) phiThetaPsi.getY())  // roll -> theta
                       .setPsi((float) phiThetaPsi.getZ()); // pitch -> psi
            
            Vector3Float linearVel = entityStatePdu_1.getEntityLinearVelocity();
            linearVel.setX(lv.getX());
        
        } catch (UnsupportedCotVersionException | JAXBException ex) {
            Logger.getLogger(CoTMsgParserDisRunner.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        getDisChannel().sendSinglePdu(entityStatePdu_1);
    }

    /**
     * @return the disChannel
     */
    public DisChannel getDisChannel() {
        return disChannel;
    }
    
    private static CoTMsgParserDisRunner thisProgram;
  
    /**
     * Main method is first executed when a program instance is loaded.
     * @see <a href="https://docs.oracle.com/javase/tutorial/getStarted/application/index.html">Java Tutorials: A Closer Look at the "Hello World!" Application</a>
     * @param args command-line parameters: network address and port.
     *    Command-line arguments are an array of optional String parameters that are passed from execution environment during invocation
     */
    public static void main(String[] args)
    {
        thisProgram = new CoTMsgParserDisRunner("test constructor"); // create instance of self within static main() method
        
        thisProgram.getDisChannel().printlnTRACE("main() started...");
        
        thisProgram.handleArguments(args); // process any command-line invocation arguments

        thisProgram.runSimulationLoops(); // ... your simulation execution code goes in there ...
        
        thisProgram.getDisChannel().tearDownNetworkInterface(); // make sure no processes are left lingering
        
        thisProgram.getDisChannel().printlnTRACE("complete."); // report successful completion
    }
    
} // end class CoTMsgParserDisRunner
