/*
Copyright (c) 1995-2024 held by the author(s).  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials provided with the
      distribution.
    * Neither the names of the Naval Postgraduate School (NPS)
      Modeling, Virtual Environments and Simulation (MOVES) Institute
      (http://www.nps.edu and https://my.nps.edu/web/moves)
      nor the names of its contributors may be used to endorse or
      promote products derived from this software without specific
      prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
package edu.nps.moves.cot.file;

import edu.nps.moves.cot.dis.CoTMsgParserDisRunner;
import edu.nps.moves.util.LogUtils;

import java.io.IOException;
import java.io.Reader;
import java.nio.CharBuffer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import mil.army.usareur.g3.mcsd.CoTExample.CustomCoTparser;
import mil.army.usareur.g3.mcsd.CoTUtils.CoTparser;

import org.apache.logging.log4j.Logger;

/** CoT message reader demonstration. This will first read a CoT PLI message.
 * Two parsers will process the read message. One will display to console as 
 * a JSON object. The other will generate DIS PDUs populated with CoT PLI data 
 * and record to a PDU log.
 * 
 * @author <a href="mailto:tdnorbra@nps.edu?subject=edu.nps.moves.cot.file.CoTFileReader">Terry Norbraten, NPS MOVES</a>
 */
public class CoTFileReader {
    
    private static final Logger LOG = LogUtils.getLogger(CoTFileReader.class);
    private static final String USAGE_MSG = "[relative path to CoT file]";
    
    private final List<CoTparser> cotParsers;
    
    private boolean customParsersSet;
    
    /** Default constructor */
    public CoTFileReader() {
        customParsersSet = false;
        cotParsers = new ArrayList<>();
    }
    
    /**
     * Add a customized CoT parser to be used for parsing the CoT events. This
     * customized parser contains the developers specific implementation and
     * handling of CoT events.
     *
     * @param cp an implementation of the CoTparser
     */
    public void addCoTparser(CoTparser cp) {
        this.cotParsers.add(cp);
        customParsersSet = true;
    }
    
    /** Reads and processes CoT PLI files
     * 
     * @param file the CoT PLI file to read and process
     */
    public void readAndProcessCoTFile(Path file) {
        String cotXml;
        
        if (!this.customParsersSet)
            cotParsers.add(new CoTparser()); // Create a default Cursor on Target parser
        
        // Convert the byte stream to characters
        CharBuffer cbuff = CharBuffer.allocate((int) file.toFile().length());
        try (Reader reader = Files.newBufferedReader(file)) {
            reader.read(cbuff);
        } catch (IOException ex) {
            LOG.error(ex);
        }
        cotXml = cbuff.flip().toString(); // don't forget to flip the buffer!!
        
        for (CoTparser p : cotParsers)
            p.coTeventHandler(cotXml);
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        
        Path p;
        if (args.length != USAGE_MSG.length()) {
            p = Path.of("src/test/resources/MT-5_Alert_Example_without_Alert_ObjectCutout_Image.cot");
            LOG.warn("Arguments must be: {}", USAGE_MSG);
            LOG.info("Using default CoT file: {}", p);
        } else
            p = Path.of(args[0]);
        
        CustomCoTparser cp = new CustomCoTparser();
        CoTMsgParserDisRunner disr = new CoTMsgParserDisRunner("CoT PLI data DIS transmitter");
        
        CoTFileReader r = new CoTFileReader();
        r.addCoTparser(cp);
        r.addCoTparser(disr);
        r.readAndProcessCoTFile(p);
    }

} // end class CoTFileReader
