/*
Copyright (c) 1995-2024 held by the author(s).  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer
      in the documentation and/or other materials provided with the
      distribution.
    * Neither the names of the Naval Postgraduate School (NPS)
      Modeling, Virtual Environments and Simulation (MOVES) Institute
      (http://www.nps.edu and https://my.nps.edu/web/moves)
      nor the names of its contributors may be used to endorse or
      promote products derived from this software without specific
      prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/
package edu.nps.moves.cot.net;

import edu.nps.moves.cot.dis.CoTMsgParserDisRunner;
import edu.nps.moves.util.LogUtils;

import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Arrays;
import java.util.concurrent.TimeUnit;

import mil.army.usareur.g3.mcsd.CoTUtils.*;
import mil.army.usareur.g3.mcsd.CoTExample.CustomCoTparser;

import org.apache.logging.log4j.Logger;

/** CoT message receiver (UPD, TCP) demonstration. This will start a CoT PLI
 * message generator which will transmit CoT (XML) over a local network Socket.
 * Two parsers will process the received message. One will display to console as
 * a JSON object. The other will generate DIS PDUs populated with CoT PLI data
 * and record to a PDU log.
 *
 * Based off of mil.army.usareur.g3.mcsd.CoTExample.CoTreceiver
 *
 * System.in simulation based on: https://stackoverflow.com/questions/1647907/junit-how-to-simulate-system-in-testing
 *
 * @author <a href="mailto:tdnorbra@nps.edu?subject=edu.nps.moves.main.CoTMsgReceiverDemo">Terry Norbraten, NPS MOVES</a>
 */
public class CoTMsgReceiverDemo {

    private static final Logger LOG = LogUtils.getLogger(CoTMsgReceiverDemo.class);

    /**
     * Utility method that sleeps for a period of time
     *
     * @param seconds integer that defines how many seconds to sleep
     */
    private static void sleep(long seconds) {
        try {
            TimeUnit.SECONDS.sleep(seconds);
        } catch (InterruptedException ex) {
            LOG.error(ex);
        }
    }//sleep

    static final String[] ARGS_DEFAULT = {"--performanceTestSendCoT", "udp", "localhost", "9999", "1", "1"};

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args)
    {

        LOG.info("Input args:");
        for (String arg:args)
        {
             LOG.info(arg);
        }
        System.out.println();

        boolean isTcp = false;
        if (args.length > 0)
        {
            isTcp = args[1].equalsIgnoreCase("tcp");
        }
        else
        {
            LOG.info("Invoked without parameter args, using default args {} ", Arrays.toString(ARGS_DEFAULT));
        }

        // - Create a custom CoTparser class. See CustomCoTparser.java.
        // - Override the coTeventHandler() to add your custom CoT data handling.
        // - Invoke the customized parser:
        CustomCoTparser cp = new CustomCoTparser();
        CoTMsgParserDisRunner disr = new CoTMsgParserDisRunner("CoT PLI data DIS transmitter");

        // Manage several UDP and TCP listeners
        // - Create a CoT connector
        CoTconnectors connectors = new CoTconnectors();

        // - Assign custom parsers to be used by the listeners. Note that the
        //   CoT custom parser must be assigned before the listeners are started
        connectors.addCoTparser(cp);
        connectors.addCoTparser(disr);

        // - Common usage example:
        System.out.println();
        LOG.info("Starting listener pool\n");

        if (isTcp)
        {
            connectors.addTcpListener(CoTtcpListener.DEFAULT_COT_TCP_PORT);
            connectors.startListener (CoTtcpListener.DEFAULT_COT_TCP_PORT);
        }
        else
        {
            connectors.addUdpListener(CoTudpListener.DEFAULT_COT_UDP_PORT); // default packet size 1024
            connectors.startListener (CoTudpListener.DEFAULT_COT_UDP_PORT);
        }

        PipedOutputStream pipedOutputStream = null;

        // Redirect an output stream to be read by an input stream in order to
        // stop the TCP listener
        try {
            pipedOutputStream = new PipedOutputStream();
            System.setIn(new PipedInputStream(pipedOutputStream));
        }
        catch (IOException ioe)
        {
            LOG.error(ioe);
        }

        // Start the CoT message generator/server
        Runnable r = () -> {
            // odd invocation avoids resetting empty args array, which can provoke following obscure Java error
            //      error: local variables referenced from a lambda expression must be final or effectively final
            //      com.aciedge.cotplitool.Main.main(args);
            if  (args.length > 0)
                com.aciedge.cotplitool.Main.main(args);
            else
                com.aciedge.cotplitool.Main.main(ARGS_DEFAULT);
        };
        Thread t = new Thread(r);
        t.setDaemon(true);
        t.start();

        // TODO: can make this another input argument
        sleep(5L); // default to 5 seconds of operation

        System.out.println();
        LOG.info("Stopping listener pool\n");

        if (isTcp)
            connectors.stopListener(CoTtcpListener.DEFAULT_COT_TCP_PORT);
        else
            connectors.stopListener(CoTudpListener.DEFAULT_COT_UDP_PORT);

        // Issue the CoT message generator/server quit/exit command
        try {
            if (pipedOutputStream != null)
                pipedOutputStream.write("quit\n".getBytes("utf-8"));
        } catch (IOException ex) {
            LOG.error(ex);
        }

        // Shutdown the DIS listerner
        disr.getDisChannel().tearDownNetworkInterface();

    }// main

} // end class CoTMsgReceiverDemo
