/*
 * The MIT License
 *
 * Copyright 2018 Martin Dudel.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package mil.army.usareur.g3.mcsd.CoTExample;

import com.aciedge.cotplitool.Constants;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Reader;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.xml.bind.JAXBException;

import mil.af.cursorOnTarget.CotDetail;
import mil.af.cursorOnTarget.CotEvent;
import mil.af.cursorOnTarget.CotPoint;
import mil.af.cursorOnTarget.DetailSubelement;
import mil.af.cursorOnTarget.UnsupportedCotVersionException;

import mil.army.usareur.g3.mcsd.CoTUtils.CoTparser;
import mil.army.usareur.g3.mcsd.CotBindings.Track;

import org.jdom2.Document;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;
import org.jdom2.input.sax.XMLReaderJDOMFactory;
import org.jdom2.input.sax.XMLReaderXSDFactory;
import org.jdom2.output.Format;
import org.jdom2.output.XMLOutputter;

/** Output received CoT XML messages to file and transmit DIS
 * 
 * @author martin.c.dudel.civ@mail.mil
 * @author <a href="mailto:tdnorbra@nps.edu?subject=mil.army.usareur.g3.mcsd.CoTExample.CustomCoTparser">Terry Norbraten, NPS MOVES</a>
 */
public class CustomCoTparser extends CoTparser {
    
    /** Schema directory */
    private static final String SCHEMAS_DIR = "lib/xsd";
    
    /** CoT Base Event Schema path */
    private static final String COT_BASE_EVENT_SCHEMA = SCHEMAS_DIR + "/CoT Base-Event Schema  (PUBLIC RELEASE).xsd";
    
    /**
     * Default format for LocaDateTime objects: "E-yyyyMMdd-HHmmss-SSS"
     */
    private static final String DTF_FORMAT = "E-yyyyMMdd-HHmmss-SSS";
    private static final DateTimeFormatter DATE_TIME_GROUP_FORMAT = DateTimeFormatter.ofPattern(DTF_FORMAT);
    
    /** Our SAXBuilder */
    private SAXBuilder builder;
    private final XMLOutputter xmlOutput;
    private File cotXmlFile;
    
    public CustomCoTparser() {
        super();
        xmlOutput = new XMLOutputter(Format.getPrettyFormat());
        
        Set<File> files = null;
        // borrowed from: https://www.baeldung.com/java-list-directory-files
        try (Stream<Path> stream = Files.list(Paths.get(SCHEMAS_DIR))) {
            files = stream
                    .filter(file -> !Files.isDirectory(file))
                    .map(Path::toFile)
                    .collect(Collectors.toSet());
        } catch (IOException ex) {
            Logger.getLogger(CustomCoTparser.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        if (files != null) {
//            File[] filez = new File[files.size()];
        
            try {
                XMLReaderJDOMFactory schemafac = new XMLReaderXSDFactory(COT_BASE_EVENT_SCHEMA);
                
                // Not working quite right, not all schema grammar locations are loaded/listed
//                XMLReaderJDOMFactory schemafac = new XMLReaderXSDFactory(files.toArray(filez));
                builder = new SAXBuilder(schemafac);
            } catch (JDOMException ex) {
                Logger.getLogger(CustomCoTparser.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }

    @Override
    public void coTeventHandler(String xmlEvent) {
        //System.out.println(Constants.CUSTOM_HANDLER_COT + " XML EVENT\n-->" + xmlEvent + "<--");
        System.out.println(Constants.CUSTOM_HANDLER_COT_JSON + " EVENT\n" + this.parseToJSON(xmlEvent.trim()));
        // Get the CoT event
        StringReader xmlReader;
        Track track;

        try {
            // Trim the xml string to avoid the following error: "Content is not allowed in trailing section"
            CotEvent cotEvent = CotEvent.parse(xmlEvent.trim());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " EVENT UID: " + cotEvent.getUID());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " EVENT TYPE: " + cotEvent.getType());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " EVENT 2525B: " + this.convertTypeTo2525B(cotEvent.getType()));
            System.out.println(Constants.CUSTOM_HANDLER_COT + " EVENT DESCRIPTION: " + this.getCotTypeDescription(cotEvent.getType()));
            
            CotPoint point = cotEvent.getPoint();
            System.out.println(Constants.CUSTOM_HANDLER_COT + " POINT LAT: " + point.getLatitude());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " POINT LON: " + point.getLongitude());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " POINT CE: " + point.getCircularError());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " POINT LE: " + point.getLinearError());
            System.out.println(Constants.CUSTOM_HANDLER_COT + " POINT HAE: " + point.getHAE());
            
            CotDetail detail = cotEvent.getDetail();
            for (DetailSubelement subElement : detail) {
                xmlReader = new StringReader(subElement.toXml());
                if ("track".equalsIgnoreCase(subElement.getName())) {
                    track = (Track) jaxbUnmarshallerTrack.unmarshal(xmlReader);
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   Course: " + track.getCourse());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   ECourse: " + track.getECourse());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   Slope: " + track.getSlope());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   ESlope: " + track.getESlope());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   Speed: " + track.getSpeed());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   ESpeed: " + track.getESpeed());
                    System.out.println(Constants.CUSTOM_HANDLER_COT + " DETAIL " + subElement.getName() + "   Version: " + track.getVersion());
                }//track
                xmlReader.close();
            }
            System.out.println();
            
            writeToFile(cotEvent);
        } catch (UnsupportedCotVersionException | JAXBException ex) {
            Logger.getLogger(CustomCoTparser.class.getName()).log(Level.SEVERE, null, ex);
        }
    }//coTeventHandler
    
    /** Handle to the last CoT file logged
     * 
     * @return a handle to the last file logged
     */
    public File getCotXmlFile() {
        return cotXmlFile;
    }
    
    /**
     * Write out a received CoT XML message to file with its timestamp in the filename
     * @param cotEvent a received CoT message
     */
    private void writeToFile(CotEvent cotEvent) {
        File dir = new File("logs");
        if (!dir.exists())
            dir.mkdir();
        
        LocalDateTime ldt = new java.sql.Timestamp(cotEvent.getStart().getTime()).toLocalDateTime();
        String dtgStamp = DATE_TIME_GROUP_FORMAT.format(ldt);
        
        // File dump directory per start time
        File dir2 = new File(dir, "CoTMsgGrp-" + dtgStamp);
        if (!dir2.exists())
            dir2.mkdir();
        
        ldt = new java.sql.Timestamp(cotEvent.getTime().getTime()).toLocalDateTime();
        dtgStamp = DATE_TIME_GROUP_FORMAT.format(ldt);
        cotXmlFile = new File(dir2, "CoTEvt-" + dtgStamp + ".xml");
        
        Document cotMsg = validateCotMsg(cotEvent.toXml());
        try (OutputStream out = new FileOutputStream(cotXmlFile)) {
            xmlOutput.output(cotMsg, out);
        } catch (IOException ex) {
            Logger.getLogger(CustomCoTparser.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    /**
     * Takes the received CoT XML message and inputs into a JDOM Document as 
     * well as validate against a given CoT schema
     * 
     * @param msg a received XML message
     * @return a JDOM Document of the message
     */
    private Document validateCotMsg(String msg) {
        Document cotMsg = null;
        try (Reader reader = new StringReader(msg)) {
            cotMsg = builder.build(reader);
        } catch (JDOMException | IOException ex) {
            Logger.getLogger(CustomCoTparser.class.getName()).log(Level.SEVERE, null, ex);
        }
        return cotMsg;
    }

} // end class CustomCoTparser
