/*
 * The MIT License
 *
 * Copyright 2018 Martin Dudel.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package mil.army.usareur.g3.mcsd.CoTUtils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/** TCP CoT msg listener
 *
 * @author martin.c.dudel.civ@mail.mil
 * @author <a href="mailto:tdnorbra@nps.edu?subject=mil.army.usareur.g3.mcsd.CoTUtils.CoTtcpListener">Terry Norbraten, NPS MOVES</a>
 */
public class CoTtcpListener extends Thread {
    
    public static final String TCP_LISTENER = "CoT TCP Listener";
    public static final int DEFAULT_COT_TCP_PORT = 9998;

    private ServerSocket cotSocket;
    private int cotTcpPort;
    private boolean debug = false;
    private volatile boolean runFlag = true;
    private final List<CoTparser> cotParsers;
    private boolean customParsersSet = false;
    private BufferedReader reader;

    /**
     * Create a CoT TCP socket listener on the default port of 9998
     */
    public CoTtcpListener() {
        this(DEFAULT_COT_TCP_PORT);
    }//CotTcpListener

    /**
     * Create a CoT TCP socket listener on the desired port.
     *
     * @param port an integer port number
     */
    public CoTtcpListener(int port) {
        this(port, false);
    }//CotTcpListener(int port)

    /**
     * Create a CoT TCP listener on the desired port and set the debug statement
     * option. Debug statements are not printed by default.
     *
     * @param port an integer port number
     * @param debug a boolean indicating whether to print out debug statements
     */
    public CoTtcpListener(int port, boolean debug) {
        cotTcpPort = port;
        this.debug = debug;
        debugToConsole(TCP_LISTENER + " port: " + cotTcpPort);
        try {
            cotSocket = new ServerSocket(cotTcpPort);
        } catch (IOException ex) {
            Logger.getLogger(CoTtcpListener.class.getName()).log(Level.SEVERE, null, ex);
        }
        cotParsers = new ArrayList<>();
    }//CotTcpListener(int port)

    /**
     * Add a customized CoT parser to be used for parsing the CoT events. This
     * customized parser contains the developers specific implementation and
     * handling of CoT events.
     *
     * @param cp an implementation of the CoTparser
     */
    public void addCoTparser(CoTparser cp) {
        this.cotParsers.add(cp);
        customParsersSet = true;
    }

    private void debugToConsole(String msg) {
        if (this.debug) {
            System.out.println(msg);
        }
    }//debugOut

    /**
     * Stops the listener. Note that when listener threads are stopped they can
     * not be restarted. The listener object must be reconstructed.
     */
    public void stopThread() {
        debugToConsole(TCP_LISTENER + " received stop request, port: " + cotTcpPort);
        this.runFlag = false;
        try {
            cotSocket.close();
        } catch (IOException ex) {
            Logger.getLogger(CoTtcpListener.class.getName()).log(Level.SEVERE, null, ex);
        }
    }//stopThread

    /**
     *
     * @return an integer representing the port the socket listener uses
     */
    public int getPort() {
        return this.cotTcpPort;
    }//getPort

    @Override
    public void run() {
        Socket accept = null;
        String cotLine;
        StringBuilder cotMessage = new StringBuilder();
        debugToConsole(TCP_LISTENER + " thread started, port: " + cotTcpPort);
        if (!this.customParsersSet) {
            cotParsers.add(new CoTparser()); // Create a default Cursor on Target parser
        }
        while (this.runFlag) {
            try {
                accept = cotSocket.accept();
                reader = new BufferedReader(new InputStreamReader(accept.getInputStream()));

                while ((cotLine = reader.readLine()) != null) {
                    cotMessage.append(cotLine);
                }
                
                // NOTE: All XML messages will come over in one big string, so,
                // parse out the individual messages and process each
                String[] cotMessages = cotMessage.toString().split("<\\?xml.*?\\?>");
                
                for (String cotXml : cotMessages) {
                    
                    // XML header was stripped out
                    if (cotXml.isBlank())
                        continue;
                
                    debugToConsole("TCP ========== Begin message:\n" + com.aciedge.cotplitool.Constants.XML_HEADER + cotXml + "\nTCP ========== End message");
                    for (CoTparser p : cotParsers) {
                        if (this.debug)
                            p.dumpCoTevent(cotXml);

                        // The CoT message is handled here                
                        p.coTeventHandler(cotXml);
                    }
                }
                // Reset the StringBuilder to hold the next TCP CoT message
                cotMessage.setLength(0);
            } catch (IOException e) {
                this.runFlag = false;
                System.err.println(e.getMessage() + "\n" + e.toString());
                System.err.println(TCP_LISTENER + " thread stopped, port: " + cotTcpPort);
            } finally {
                try {
                    if (accept != null)
                        accept.close();
                } catch (IOException ex) {
                    Logger.getLogger(CoTtcpListener.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }//while
    }//run    
}// class CoTtcpListener
