/**
 * Copyright (c) 2008-2023, MOVES Institute, Naval Postgraduate School (NPS). All rights reserved.
 * This work is provided under a BSD-style open-source license, see project
 * <a href="https://savage.nps.edu/opendis7-java/license.html" target="_blank">license.html</a> and <a href="https://savage.nps.edu/opendis7-java/license.txt" target="_blank">license.txt</a>
 */
package edu.nps.moves.dis7.utilities;

import edu.nps.moves.dis7.utilities.stream.PduRecorder;
import edu.nps.moves.dis7.utilities.stream.PduRecorderIEEE;

/**
 * DisChannel integrates multiple utility capabilities to handle most networking and entity-management tasks.
 * Provides a simplified interface wrapping DisThreadedNetworkInterface, PduRecorderIEEE, and SimulationManager
 * for programs connecting to OpenDis7 communications.
 * 
 * Based on the original <a href="https://github.com/open-dis/opendis7-java" target="_blank">edu.nps.moves.dis7.utilities.DisChannel</a>
 * Modified to invoke the PduRecorderIEEE which records PLAINTEXT PDUs without headers/footers/comments
 * 
 * <br>
 * @see <a href="https://gitlab.nps.edu/Savage/NetworkedGraphicsMV3500/-/blob/master/specifications/README.md">MV3500 Distributed Simulation Fundamentals, Specification Documents, IEEE and SISO</a>
 * @see <a href="https://ieeexplore.ieee.org/document/6387564">1278.1-2012. IEEE Standard for Distributed Interactive Simulation (DIS) - Application Protocols</a>
 * 
 * @author <a href="mailto:tdnorbra@nps.edu?subject=edu.nps.moves.dis7.utilities.DisChannelIEEE">Terry D. Norbraten</a>
 */
public class DisChannelIEEE extends DisChannel
{
    private static String PDU_ENCODING;
    
    /** Base constructor */
    public DisChannelIEEE()
    {
        super();
    }
    
    /** Constructor with new descriptor
     * @param initialDescriptor descriptor for this instance 
     */
    public DisChannelIEEE(String initialDescriptor)
    {
        super(initialDescriptor);
    }
    
    /** Constructor with new verboseness
     * @param verbosenessDisNetworkInterface  whether channel is initially verbose
     * @param verbosenessPduRecorder      whether PduRecorder is initially verbose 
     */
    public DisChannelIEEE(boolean verbosenessDisNetworkInterface, boolean verbosenessPduRecorder)
    {
        super(verbosenessDisNetworkInterface, verbosenessPduRecorder);
    }
    
    /** Constructor with new descriptor, verboseness
     * @param initialDescriptor descriptor for this instance
     * @param verbosenessDisNetworkInterface  whether channel is initially verbose
     * @param verbosenessPduRecorder whether PduRecorder is initially verbose  
     */
    public DisChannelIEEE(String initialDescriptor, boolean verbosenessDisNetworkInterface, boolean verbosenessPduRecorder)
    {
        super(initialDescriptor, verbosenessDisNetworkInterface, verbosenessPduRecorder);
    }
    
    @Override
    public synchronized void setUpNetworkInterface() 
    {
        if (disNetworkInterface != null)
        {
            printlnTRACE("*** Warning: setUpNetworkInterface() has already created disNetworkInterface, second invocation ignored");
            return;
        }
        
        String pduLogOutputDirectory = DEFAULT_PDULOG_OUTPUT_DIRECTORY;
        printlnTRACE("Beginning pdu save to directory " + pduLogOutputDirectory);
        pduRecorder = new PduRecorderIEEE(pduLogOutputDirectory, getNetworkAddress(), getNetworkPort()); // assumes save
        pduRecorder.setEncodingPduLog(getPDU_ENCODING());
        
        // Here is where we need plain text PDU output in CSV with no headers/footers/comments, etc
        if (getPDU_ENCODING().equals(PduRecorder.ENCODING_PLAINTEXT))
           pduRecorder.setIncludeHeaders(false); // was set true in above method call
        
        pduRecorder.setDescriptor(descriptor); // also sets for the DisThreadedNetworkInterface
        pduRecorder.setLogFileName(descriptor + "_" + getPDU_ENCODING() + "_" + PduRecorder.DEFAULT_FILE_NAME);
        pduRecorder.setVerbose(isVerboseDisNetworkInterface()); // either sending, receiving or both. Also sets for the DisThreadedNetworkInterface
        pduRecorder.start(); // begin running which also starts a DisThreadedNetworkInterface instance with the given network address and port
        
        disNetworkInterface = pduRecorder.getDisThreadedNetworkInterface();
        if (isVerbosePduRecorder())
            printlnTRACE("Network confirmation:" + " address=" + getDisNetworkInterface().getAddress() + //  disNetworkInterface.getMulticastGroup() +
                                                      " port=" + getDisNetworkInterface().getPort());    // + disNetworkInterface.getDisPort());
    }

    /**
     * @return the PDU_ENCODING
     */
    public static String getPDU_ENCODING() {
        return PDU_ENCODING;
    }

    /**
     * @param PDU_ENCODING the PDU_ENCODING to set
     */
    public static void setPDU_ENCODING(String PDU_ENCODING) {
        DisChannelIEEE.PDU_ENCODING = PDU_ENCODING;
    }
    
} // end class DisChannelIEEE
