/**
 * Copyright (c) 2008-2023, MOVES Institute, Naval Postgraduate School (NPS). All rights reserved.
 * This work is provided under a BSD-style open-source license, see project
 * <a href="https://savage.nps.edu/opendis7-java/license.html" target="_blank">license.html</a> and <a href="https://savage.nps.edu/opendis7-java/license.txt" target="_blank">license.txt</a>
 */
package edu.nps.moves.dis7.utilities.stream;

import edu.nps.moves.dis7.enumerations.DisPduType;
import edu.nps.moves.dis7.pdus.Pdu;
import edu.nps.moves.dis7.utilities.DisThreadedNetworkInterface;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.util.Arrays;
 
/** Utility to save PDUs received over the network to disk, greatly simplifying the capture of DIS streams by applications.
 * 
 * Example <code>main()</code> self-test response shown in log file, example results also included.
 * @see <a href="https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestLog.txt">https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestLog.txt</a>
 * @see <a href="https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_BASE64_PduCaptureLog.dislog">https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_BASE64_PduCaptureLog.dislog</a>
 * @see <a href="https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_BINARY_PduCaptureLog.dislog">https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_BINARY_PduCaptureLog.dislog</a>
 * @see <a href="https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_PLAINTEXT_PduCaptureLog.dislog">https://github.com/open-dis/open-dis7-java/blob/master/src/edu/nps/moves/dis7/utilities/stream/PduRecorderSelfTestENCODING_PLAINTEXT_PduCaptureLog.dislog</a>
 * 
 * Based on the original <a href="https://github.com/open-dis/opendis7-java" target="_blank">edu.nps.moves.dis7.utilities.stream.PduRecorder</a>
 * Modified to record PLAINTEXT PDUs without headers/footers/comments
 * 
 * @author Don Brutzman, brutzman@nps.edu
 * @author Mike Bailey, jmbailey@nps.edu
 * @author <a href="mailto:tdnorbra@nps.edu?subject=edu.nps.moves.dis7.utilities.stream.PduRecorderIEEE">Terry D. Norbraten</a>
 */
public class PduRecorderIEEE extends PduRecorder // implements PduReceiver
{ 
    /**
     * Default constructor that uses default values for output directory, DIS address and port.
     * Each instance must invoke start() to begin operations, pause() to suspend operations, 
     * resume() to continue operations, and stop() to terminate operations.
     */
    public PduRecorderIEEE()
    {
        this(OUTPUT_DIRECTORY_DEFAULT);
    }

    /**
     * Constructor to let the use specify an output directory.
     * Uses default values for multicast address and port.
     * Each instance must invoke start() to begin operations, pause() to suspend operations, 
     * resume() to continue operations, and stop() to terminate operations.
     * 
     * @param initialOutputDirectory the directory to write log files to
     */
    public PduRecorderIEEE(String initialOutputDirectory)
    {
        this(initialOutputDirectory, DEFAULT_DIS_ADDRESS, DEFAULT_DIS_PORT);
    }

    /** Constructor to let the user specify all required parameters.
     * Each instance must invoke start() to begin operations, pause() to suspend operations, 
     * resume() to continue operations, and stop() to terminate operations..
     * 
     * @param initialOutputDirectory local path for directory where the log files are written
     * @param initialAddress multicast group address to receive data from (TODO allow unicast UDP)
     * @param initialPort UDP port to listen for data
     */
    public PduRecorderIEEE(String initialOutputDirectory, String initialAddress, int initialPort)
    {
        super(initialOutputDirectory, initialAddress, initialPort);
        setIncludeHeaders(false);
        setReadableTimeStamp(false);
        setDescriptor(getClass().getSimpleName());
    }

    /** receivePdu from DIS data stream, invoked via callback from DisThreadedNetworkInterface.RawPduListener
     * @param newBuffer byte array for receiving data
     * @param newLength length of byte array
     */
    @Override
    public void receivePdu(byte[] newBuffer, int newLength)
    {
      if(!isRunning())
        return; // thread operations no longer in progress, ignore this received PDU

      byte[]    byteBufferSized = Arrays.copyOf(newBuffer, newLength);
      
      switch (encodingPduLog)
      {
          case ENCODING_BINARY:
              // diagnostics can go here, TODO is any processing needed?
              break;

          case ENCODING_BASE64:
              sb.append(base64Encoder.encodeToString(byteBufferSized));
              break;

          case ENCODING_PLAINTEXT:
              
              // No timestamp in the beginning, just pure PDU contents
              // PDU contents, remove square brackets to end up with pure CSV
              sb.append(Arrays.toString(byteBufferSized).replace(" ", "").replace("[","").replace("]",""));
              break;
              
          default:
              if (ENCODING_OPTIONS_LIST.contains(encodingPduLog))
                 System.err.println ("Encoding " + encodingPduLog + " not supported");
              else
                 System.err.println ("Encoding " + encodingPduLog + " not recognized");
              break;
      }
      try
      {
        if (encodingPduLog.equals(ENCODING_BINARY))
        {
            // https://stackoverflow.com/questions/4931854/converting-char-array-into-byte-array-and-back-again
            Charset    charset = Charset.forName("ISO-8859-1"); // Must be this to get the proper encoding for DFDL processing
//            Charset    charset = Charset.forName("UTF_8");
//            Charset    charset = Charset.forName("US-ASCII");
            CharBuffer charBufferSized = charset.decode(ByteBuffer.wrap(byteBufferSized));
            logFileWriter.write(charBufferSized.array());
        }
        else
        {
            logFileWriter.write(sb.toString());
            ((PrintWriter)logFileWriter).println();
        }
      }
      catch (IOException ex)
      {    
          throw new RuntimeException("Fatal exception writing DIS log file in " + getDescriptor() + " thread, encodingPduLog=" + encodingPduLog + ": " + ex);
      }
      pduCount += 1;
//      if (false) // debug
//          System.out.println(TRACE_PREFIX + " + getDescriptor() + ": pduCount="+ pduCount);

      sb.setLength(0);
    }

    /**
     * Entry point invocation to facilitate developmental testing, runs selfTest() method.
     *
     * @param args none supported, TODO offer path/filename
     */
    public static void main(String[] args)
    {
        PduRecorder pduRecorder = new PduRecorderIEEE(); // instantiates an instance object of this class,
        pduRecorder.selfTest(args); // and tests it.  Results shown in .dislog files.
    }
    
  /** This selfTest() method saves PDU output logs to assigned directory using all supported encodings. 
   *  Further checking can be accomplished by separately Invoking the edu.nps.moves.dis7.examples.PduReaderPlayer
   *  which will playback all logs written to that log directory.
   * 
   * @param args none supported, TODO offer path/filename
   */
  @Override
  public synchronized void selfTest(String[] args)
  {
    System.out.println(getDescriptor() +  " main() performs self-test by sending full set of PDUs");
    
    PduRecorder pduRecorder;
    DisThreadedNetworkInterface disNetworkInterface;
    DisPduType allPDUTypesArray[] = DisPduType.values();
    System.out.println(getDescriptor() +  " allPDUTypesArray created, length=" + allPDUTypesArray.length + " ...");
    System.out.flush(); // ensure all output sent
    
    for (String currentEncoding : ENCODING_OPTIONS_LIST)
    {
        pduRecorder = new PduRecorderIEEE(); // default address, port, output directory path
        
        System.out.println("=================================================");
        System.out.println("Test " + getDescriptor() + " encoding " + currentEncoding);
        pduRecorder.setEncodingPduLog(currentEncoding);
        
        // Here is where we need plain text PDU output in CSV with no headers/footers/comments, etc
        if (pduRecorder.getEncodingPduLog().equals(ENCODING_PLAINTEXT))
            pduRecorder.setIncludeHeaders(false); // was set true in above method call
        
        pduRecorder.setDescriptor(getDescriptor() + ".selfTest " + currentEncoding);
//      pduRecorder.setPort(1); // option to avoid listening to other PDU streams during self test
        pduRecorder.setLogFileName(getDescriptor() + ".selfTest" + "_" + currentEncoding + "_" + DEFAULT_FILE_NAME);
        pduRecorder.start();
        disNetworkInterface = pduRecorder.getDisThreadedNetworkInterface(); // must reinitialize after each begin
        System.out.println(pduRecorder.getDescriptor() + " pduRecorder started... isRunning()=" + pduRecorder.isRunning());

        for (int index=0; index < allPDUTypesArray.length; index++)
        {
          DisPduType pduTypeValue = allPDUTypesArray[index];
          if (pduTypeValue != DisPduType.OTHER) 
          {
            try {
                // These are empty PDUs, no boilerplate data marshalled
                Pdu nextPdu = pduFactory.createPdu(allPDUTypesArray[index]);
                nextPdu.setTimestamp(index * 10); // seconds
                nextPdu.setLength(nextPdu.getMarshalledSize());
//              nextPdu.getTimestamp(); // debug
                disNetworkInterface.sendPDU(nextPdu);
                
                // https://stackoverflow.com/questions/10663920/calling-thread-sleep-from-synchronized-context-in-java
                // https://stackoverflow.com/questions/1036754/difference-between-wait-vs-sleep-in-java
                wait (100L); // let send/receive threads and streams catch up // TODO consider wait() instead of sleep()
            }
            catch (InterruptedException ex) {
              System.err.println("Exception sending Pdu " + pduTypeValue + ": " + ex.getLocalizedMessage());
            }
          }
          else 
          {
              System.err.println("Found pduTypeValue=DisPduType.OTHER=" + pduTypeValue);
          }
        }
        System.out.flush();
        System.err.flush(); // ensure all output sent
        pduRecorder.stop();
        System.out.println(pduRecorder.getDescriptor() + " pduRecorder complete... isRunning()=" + pduRecorder.isRunning());
    }
    // end of loop ENCODING_OPTIONS_LIST
    System.out.println("=================================================");
  }
    
} // end class PduRecorderIEEE
